//
// Power remote 230V Relay Control
// GRA
// Créé le : 22.04.2018
// Der modif : 26.05.2018
// V1.0
// V1.1 Ajout du switch,
// V1.2 Ajout d'un "else{return};" dans le callback pour voir si cela corrige un bug quand il y a beaucoup de messages qui arrivent dans un délais court
// V1.3 Ajout du test si bouton pressé plus que 1s, reboot le ESP8266
// V1.4 Changé config pour NAS 192.168.1.4
// V1.5 Ajout ligne pour mettre en STAtion et non AP afin d'éviter les piratages, remis Rpi en attendant de réussir à faire démarrer le NR du NAS seul
// V1.6 Déplacement de l'attente de 5s en cas de déconnexion du serveur MQTT amélioration reset et reboot
// Connect to MQTT server and subscribe to topic
// Paramètres de programmations : ESP8285, 115200 bauds, 1M no SPIFFS,26/80MHz, Led = pin 2

#include <PubSubClient.h>
#include <ESP8266WiFi.h>
#define LED_PIN 2
#define RELAY_PIN 4
#define BUTTON_PIN 0
#define ADD1 13
#define ADD2 12
#define ADD3 14
#define ADD4 5

// Ponthaux
const char* ssid = "xxx";                                   
const char* password = "yyy";
const char* server = "192.168.1.xx";

char* topic = "your/topic";    // Ajouter 0 à 15 en lisant les IO

String swVersion = "V1.6";
String hwVersion = "V1.0";

unsigned long sleepTimeS = 5;     //300 = 5mn, 600 = 10mn, 1800 = 30mn, 3600 = 1h

String payload = "";
String clientName;

char message_buff[1];   // initialise storage buffer (i haven't tested to this capacity.)
int state = 0;          // Etat de la sortie
int address = 0;
int interruption = 0;

long lastReconnectAttempt = 0;

void callback(char* topic, byte* payload, unsigned int length);

WiFiClient wifiClient;
PubSubClient client(server, 1883, callback, wifiClient);

///////////////////////////////////////////////////////////////
// Convert Mac adress to string
///////////////////////////////////////////////////////////////
String macToStr(const uint8_t* mac)
{
  String result;
  for (int i = 0; i < 6; ++i) {
    result += String(mac[i], 16);
    if (i < 5)
      result += ':';
  }
  return result;
}

///////////////////////////////////////////////////////////////
// Connect to wifi
///////////////////////////////////////////////////////////////
void connectwifi()
{
  int cnt = 0;
  Serial.print("Connecting to ");
  Serial.print(ssid);
  
  WiFi.mode(WIFI_STA);          // 26.05.2018 Mettre en station et non AP pour ne pas être piratable
  WiFi.begin(ssid, password);
  //WiFi.config(ip, gateway, subnet);
  while (WiFi.status() != WL_CONNECTED)
  {
    digitalWrite(LED_PIN, LOW);
    delay(50);
    
    Serial.print(".");
    cnt += 1;
    digitalWrite(LED_PIN, HIGH);
    if(cnt>=240)
    {
      digitalWrite(LED_PIN, LOW);
      delay(1000);
      ESP.deepSleep(1000);
    }
  }
  
  Serial.println("WiFi connected");  
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
  // Generate client name based on MAC address
  clientName += "PowerRemote-";
  uint8_t mac[6];
  WiFi.macAddress(mac);
  clientName += macToStr(mac);

  Serial.print("Connecting to ");
  Serial.print(server);
  Serial.print(" as ");
  Serial.println(clientName);

  if (client.connect((char*) clientName.c_str()))
  {
    Serial.println("Connected to MQTT broker");
    Serial.print("Topic is: ");
    Serial.println(topic);   
    char* topicTest = "test/test-connexion";
    if (client.publish(topicTest, "Starting..."))
    {
      Serial.println("Publish successfully sent");
      client.subscribe(topic);
    }
    else
    {
      digitalWrite(LED_PIN, LOW);
      Serial.println("Publish failed");
    }
  }
  else
  {
    digitalWrite(LED_PIN, LOW);
    Serial.println("MQTT connect failed");
    Serial.println("Sleeping and try again...");
    ESP.deepSleep(10000000);
  }
  digitalWrite(LED_PIN, HIGH);
}

///////////////////////////////////////////////////////////////
// Callback
///////////////////////////////////////////////////////////////

void callback(char* topicCallback, byte* payload, unsigned int length)
{
  // Here is what i have been using to handle subscriptions. I took it as a snippet from elsewhere but i cannot credit author as i dont have reference!
  int i = 0;
  
  Serial.print("Comparaison de topics : ");
  Serial.print(topic);
  Serial.print(" et ");
  Serial.println(topicCallback);
  if (String(topic) == String(topicCallback))
    {
    Serial.println("Message arrived:  topic: " + String(topicCallback));
    Serial.println("Length: " + String(length,DEC));
   
    // create character buffer with ending null terminator (string)
    for(i=0; i<length; i++)
    {
      message_buff[i] = payload[i];
      delay(20);
    }
    message_buff[i] = '\0';
   
    String msgString = String(message_buff);
   
    Serial.println("Payload: " + msgString);

    if ((msgString == "1") && (String(topic) == String(topicCallback)))   // Allumer que si c le même topic et un état 0 ou 1
    {    // if there is a "1" published to topic () on the broker then:
      digitalWrite(LED_PIN, LOW);     // set output low
      digitalWrite(RELAY_PIN, HIGH);
      Serial.println("Switching Relay ON");
      client.publish(topic, "Power ON");
      state = 1;
    }
    else if ((msgString == "0") && (String(topic) == String(topicCallback)))
    {    // if there is a "0" published to topic () on the broker then:
      digitalWrite(LED_PIN, HIGH);     // set output high
      digitalWrite(RELAY_PIN, LOW);
      Serial.println("Switching Relay OFF");
      client.publish(topic, "Power OFF");
      state = 0;
    }
    else
    {
      return;
    }
  }
  else
  {
    Serial.println("Another topic, not for me");
  }
}
///////////////////////////////////////////////////////////////
// Reconnect to MQTT client
///////////////////////////////////////////////////////////////

boolean reconnect()
{
  Serial.println("Connection loosen, try to reconnect");
  if (client.connect("ESPClient"))
  {
    // Once connected, publish an announcement...
    client.publish("outTopic","hello world");
    // ... and resubscribe
    client.subscribe(topic);
  }
  return client.connected();
}

///////////////////////////////////////////////////////////////
// Read Address
///////////////////////////////////////////////////////////////
void readaddress()
{
  address = digitalRead(ADD1);
  //Serial.print(digitalRead(ADD1));
  //Serial.print(digitalRead(ADD2));
  //Serial.print(digitalRead(ADD3));
  //Serial.println(digitalRead(ADD4));
  address = digitalRead(ADD1);
  address = address << 1;
  address = address + digitalRead(ADD2);
  address = address << 1;
  address = address + digitalRead(ADD3);
  address = address << 1;
  address = address + digitalRead(ADD4);
  Serial.print("Address : ");
  Serial.println(address);
  //topic = String(topic) + String(address);      // Doit être entre 0 et 15
  if(address == 0) {topic = "powerremote/pr_15";}
  else if(address == 1) {topic = "powerremote/pr_14";}
  else if(address == 2) {topic = "powerremote/pr_13";}
  else if(address == 3) {topic = "powerremote/pr_12";}
  else if(address == 4) {topic = "powerremote/pr_11";}
  else if(address == 5) {topic = "powerremote/pr_10";}
  else if(address == 6) {topic = "powerremote/pr_9";}
  else if(address == 7) {topic = "powerremote/pr_8";}
  else if(address == 8) {topic = "powerremote/pr_7";}
  else if(address == 9) {topic = "powerremote/pr_6";}
  else if(address == 10) {topic = "powerremote/pr_5";}
  else if(address == 11) {topic = "powerremote/pr_4";}
  else if(address == 12) {topic = "powerremote/pr_3";}
  else if(address == 13) {topic = "powerremote/pr_2";}
  else if(address == 14) {topic = "powerremote/pr_1";}
  else if(address == 15) {topic = "powerremote/pr_0";}
  else {topic = "error";}
  Serial.print("Topic used depending of address : ");
  Serial.println(topic);
}
///////////////////////////////////////////////////////////////
// buttonInterrupt
///////////////////////////////////////////////////////////////
void buttonInterrupt()
{
  interruption++;
  Serial.println("Button pressed, toggle light");
  digitalWrite(RELAY_PIN,!digitalRead(RELAY_PIN));
}

///////////////////////////////////////////////////////////////
// Setup
///////////////////////////////////////////////////////////////
void setup()
{
  pinMode(LED_PIN, OUTPUT);
  pinMode(RELAY_PIN, OUTPUT);
  pinMode(ADD1, INPUT);
  pinMode(ADD2, INPUT);
  pinMode(ADD3, INPUT);
  pinMode(ADD4, INPUT);
  pinMode(BUTTON_PIN, INPUT);
  // Configuration de l'interruption
  attachInterrupt(BUTTON_PIN, buttonInterrupt, FALLING);

  digitalWrite(LED_PIN, LOW);

  Serial.begin(74880);
  delay(100);
  Serial.println();
  Serial.println("Power Remote Switch ");
  Serial.print(swVersion);
  Serial.print(", HW:");
  Serial.println(hwVersion);
  

  readaddress();    // Lire l'adresse sur les dipswitch

  connectwifi();    // Connect wifi
  
  if (client.connected())   // Connect to broker
  {  
    if (client.publish(topic, (char*) payload.c_str(),true))
    {   
      Serial.println("Publish ok");
      client.subscribe(topic);
      digitalWrite(LED_PIN, HIGH);
    }
    else
    {    
      digitalWrite(LED_PIN, LOW);
      Serial.println("Publish failed");
    }
  }
  else
  {
    digitalWrite(LED_PIN, LOW);
    Serial.println("Try to reconnect...");
    ESP.deepSleep(1000);
  }
  for (int i=0;i<=4;i++)
  {
    digitalWrite(LED_PIN, LOW);
    delay(200);
    digitalWrite(LED_PIN, HIGH);
    delay(200);
  }
}

///////////////////////////////////////////////////////////////
// Loop
///////////////////////////////////////////////////////////////
void loop()
{
  if (!client.connected())
  {
    long now = millis();
    if (now - lastReconnectAttempt > 5000)
    {
      lastReconnectAttempt = now;
      Serial.print("Last Reconnect Attempt : ");
      Serial.println(lastReconnectAttempt);
      // Attempt to reconnect
      if (reconnect())
      {
        Serial.println("Reconnected");
        lastReconnectAttempt = 0;
      }
    }
  }
  client.loop();
  if (WiFi.status() != WL_CONNECTED)    // Si le Wifi n'est plus connecté, veille prolongée 1s, donc reset
  {
    delay(1000);
    connectwifi();
  }
  
  if(interruption >= 1)
  {
    delay(200);     // Anti-rebonds
    Serial.println("Interruption détectée");
    interruption = 0;
    detachInterrupt(BUTTON_PIN);

    state = digitalRead(RELAY_PIN);
    Serial.print("Etat du relais lu : ");
    Serial.println(state);
    char stateChar[1];
    sprintf(stateChar,"%d",state);        // Conversion du state en string pour le publier
    if(state == 0)
    {
      client.publish(topic, "Power OFF");
      digitalWrite(LED_PIN, HIGH);
    }
    else if (state == 1)
    {
      client.publish(topic, "Power ON");
      digitalWrite(LED_PIN, LOW);
    }

    int buttonPin = digitalRead(BUTTON_PIN);
    int i = 0;
    while(buttonPin == 0)       // Tant que le bouton pressé, ne pas enclancher et déclencher en boucle
    {
      wdt_reset();
      if(i>=10)
      {
        digitalWrite(LED_PIN, LOW);
        while(buttonPin == 0)       // Tant que le bouton pressé, ne pas enclancher et déclencher en boucle
        {
          wdt_reset();
          buttonPin = digitalRead(BUTTON_PIN);
        }
        digitalWrite(LED_PIN, LOW);
        ESP.deepSleep(1000);          // Si bouton pressé pendant plus de 1s, redémarrer le module
      }
      buttonPin = digitalRead(BUTTON_PIN);
      delay(100);
      i++;
    }
    delay(200);
    attachInterrupt(BUTTON_PIN, buttonInterrupt, FALLING);
  }
}

